;;;; spiffy-base.scm
;
; Copyright (c) 2007, Peter Bex
; Copyright (c) 2000-2005, Felix L. Winkelmann
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
; conditions are met:
;
;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
;     disclaimer. 
;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
;     disclaimer in the documentation and/or other materials provided with the distribution. 
;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
;     products derived from this software without specific prior written permission. 
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
; AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.
;
; Please report bugs, suggestions and ideas to the Chicken Trac
; ticket tracking system (assign tickets to user 'sjamaan'):
; http://trac.callcc.org
;

;; TODO for next release:
;; * Remove some cruft that does not belong here (nominations: load-once, escape-chars, htmlize-*)
;; * Think about possibilities for rewriting requests (use continuations to enable re-invoking the dispatch?)
;; * Think about ways to improve testing, maybe by hacking http-client (notably, sending requests without Host: line)

(require-extension utils regex posix srfi-1 srfi-13 srfi-69 http-server sendfile)

(declare 
  ;(no-bound-checks)
  (export spiffy-tcp-port spiffy-debug-mode spiffy-server-name spiffy-debug
	  spiffy-root-path spiffy-access-denied? start-server spiffy-access-file
          spiffy-vhost-map spiffy-error-log write-response-header escape-chars
	  content-transporter emit-error-message
	  write-fragment-response htmlize redirect redirect-page htmlize-with-spaces
	  strip-tags spiffy-file-type-map spiffy-index-pages current-request
	  load-once set-header! spiffy-cache-limit cache-read
	  cache-write invalidate-cache spiffy-read-block-size
	  spiffy-accept-incoming-host current-response-code current-response-headers
	  spiffy-user-name spiffy-access-log spiffy-config-file build-error-message
	  current-urlencoded-arguments current-hostname current-pathinfo
	  current-workdir current-path spiffy-file-ext-handlers spiffy-exception-handler
	  spiffy-handle-directory spiffy-not-found-handler spiffy-error-css spiffy-default-mime-type))

;;; Parameters:

(define spiffy-version 3)
(define spiffy-release 9)

;; Runtime params
(define current-request (make-parameter #f))
(define current-urlencoded-arguments (make-parameter '()))
(define current-response-headers (make-parameter '()))
(define current-response-code (make-parameter '(200 . "OK")))
(define current-workdir (make-parameter "/"))
(define current-pathinfo (make-parameter #f))
(define current-path (make-parameter #f))
(define current-hostname (make-parameter #f))

(define spiffy-tcp-port (make-parameter 8080))
(define spiffy-debug-mode (make-parameter #f))
(define spiffy-server-name
  (make-parameter (sprintf "spiffy No. ~A (Release ~A) on Chicken ~A" spiffy-version spiffy-release (chicken-version))))
(define spiffy-accept-incoming-host (make-parameter #f))
(define spiffy-user-name (make-parameter #f))
(define spiffy-access-file (make-parameter #f))
(define spiffy-config-file (make-parameter #f))
(define spiffy-access-log (make-parameter #f))
(define spiffy-error-log (make-parameter (current-error-port)))
(define spiffy-handle-directory (make-parameter (lambda (path) (http:write-error-response 403 "Forbidden"))))
(define spiffy-file-ext-handlers (make-parameter '()))
(define spiffy-not-found-handler (make-parameter #f))
(define spiffy-error-css (make-parameter "body { background-color: #ff6342; }"))
(define spiffy-exception-handler (make-parameter (lambda args (apply default-exn-handler args))))

(define spiffy-vhost-map (make-parameter `((".*" . ,(lambda (next) (next))))))
(define spiffy-root-path (make-parameter "./web"))
(define spiffy-access-denied? (make-parameter
			       (lambda (path file)
				 (or (string-prefix? "." file)  ;; dotfile?
				     (string=? ".sspx" (or (pathname-extension file) "")) ;; compiled file?
				     (any (cut string-search <> (make-pathname path file)) '("\\~" "\\.\\."))))))
(define spiffy-index-pages (make-parameter '("index.html" "index.ssp" "index.ws" "index.htm")))
(define spiffy-default-mime-type (make-parameter 'application/octet-stream))

(define spiffy-file-type-map
  (make-parameter
   '(("txt" . text/plain)
     ("xml" . text/xml)
     ("xul" . application/vnd.mozilla.xul+xml)
     ("htm" . text/html)
     ("html" . text/html)
     ("pdf" . application/pdf)
     ("css" . text/css)
     ("bmp" . image/bmp)
     ("ico" . image/x-icon)
     ("gif" . image/gif)
     ("jpg" . image/jpeg)
     ("jpeg" . image/jpeg)
     ("png" . image/png))))

(define spiffy-read-block-size (make-parameter 100000)) ; should this be smaller?

;;; Debug utils:

(define (spiffy-debug fstr . args)
  (when (spiffy-debug-mode)
    (fprintf (current-error-port) "[SPIFFY: ~?]~%~!" fstr args) ) )

(define (emit-error-message ex #!optional print-chain)
  (with-output-to-log (spiffy-error-log)
    (lambda ()
      (print-error-message ex)
      (when print-chain (print-call-chain)))))

;;; Log file handling:

(define (with-output-to-log log thunk)
  (when log
    (if (output-port? log)
	(with-output-to-port log thunk)
	(with-output-to-file log thunk append:))))

(define (log-to log fmt . rest)
  (with-output-to-log log
    (lambda ()
      (apply printf fmt rest)
      (newline))))

(define (log-error fmt . rest)
  (apply log-to (spiffy-error-log) fmt rest))

(define (log-message fmt . rest)
  (apply log-to (spiffy-access-log) fmt rest))

(http:log-hook
 (lambda (req addr)
   (when req
     (log-message "~A [~A] \"~A ~A ~A\" ~A "
		  addr
		  (seconds->string (current-seconds))
		  (http:request-method req)
		  (http:request-url req)
		  (http:request-protocol req) 
		  (alist-ref "User-Agent"
			     (http:request-attributes req)
			     string-ci=?
			     "(unknown User-Agent)")))))

;;; Output proper connnection mode:

(define (write-response-header)
  (http:write-response-header
   (current-request)   
   (car (current-response-code))
   (cdr (current-response-code))
   (current-response-headers)
   (current-output-port)))

;;; HEAD support:

(http:request-method-handler 
 'HEAD 
 (http:request-method-handler 'GET))

(http:startup-hook (constantly #f))

(define (sigint-handler sig)
  (stop-server))

(unless (eq? (build-platform) 'msvc)
  (set-signal-handler! signal/int sigint-handler))

;;; Content transporters and mime-types
(define content-transporters '())

(define (content-transporter type . proc)
  (if (pair? proc)
      (let ([a (assq type content-transporters)])
	(if a
	    (set-cdr! a (car proc))
	    (set! content-transporters (alist-cons type (car proc) content-transporters))))
      (or (and-let* ([a (assq type content-transporters)]) (cdr a))
	  default-content-transporter)))

(define (file-ext->mime-type ext)
  (alist-ref (or ext "") (spiffy-file-type-map) string-ci=? (spiffy-default-mime-type)))

(define (send-static-file fn)
  (let ([buffer (cache-read fn)])
    (if buffer
	(let ([size (##sys#size buffer)])
	  (set-header! (sprintf "Content-Length: ~A" size))
	  (write-response-header)
	  (unless (eq? 'HEAD (http:request-method (current-request)))
	    (display buffer) ) )
	(let ([size (file-size fn)])
	  (set-header! (sprintf "Content-Length: ~A" (car (string-split (number->string size) ".")))) ;; yuck!
	  (write-response-header)
	  (unless (eq? 'HEAD (http:request-method (current-request)))
	    (let ([fd (file-open fn (+ open/binary open/rdonly))])
	      (handle-exceptions exn (begin
				       (file-close fd)
				       (signal exn))
		(let ([rbs (spiffy-read-block-size)])
		  (spiffy-debug "(root) reading ~A with size ~A" fn size)
		  (if (< size rbs)
		      (let ((buf (car (file-read fd size))))
			(cache-write fn buf)
			(display buf))
		      (sendfile fd (current-output-port)))))
	      (file-close fd)))))))

(define (default-content-transporter ct fn)
  (send-static-file fn))

(http:fallback-handler
 (let ((old (http:fallback-handler)))
   (spiffy-not-found-handler (lambda () (old (current-request)))) ;; Slightly ugly, but needed so toplevel config can redefine it
   (lambda (req)
     
     (parameterize ((current-request req)
		    (current-response-code '(200 . "OK"))
		    (current-response-headers (list (cons "Content-Type" "text/html"))))
       (handle-exceptions exn
	    (parameterize ((current-response-code '(500 . "Internal server error")))
	      (emit-error-message exn)
	      ((spiffy-exception-handler) exn))
	 (dispatch-vhost old))))))

(define (normalize-path path)
  (let loop ((parts (string-split path "/"))
	     (result '()))
    (cond
     ((null? parts) (string-join (reverse result) "/" 'prefix))
     ((string=? (car parts) "..")
      (if (null? result)
	  (loop (cdr parts) result)
	  (loop (cdr parts) (cdr result))))
     ((string=? (car parts) ".")
      (loop (cdr parts) result))
     (else (loop (cdr parts) (cons (car parts) result))))))

(define (dispatch-vhost old)
  (let-values ([(path args) (http:decode-url (http:request-url (current-request)))])
    (let* ((host-header (alist-ref "Host" (http:request-attributes (current-request)) string-ci=?))
	   (host (and host-header (first (string-split host-header ":")))) ; Strip port
	   (handler (and host-header ;; make a regression test for the case of no host line.  http-client always sends a Host: line.
			 (alist-ref host (spiffy-vhost-map) (lambda (h _)
							      (if (not (regexp? h))
								  (string-match (regexp h #t) host) ;; convert to case-insensitive regex
								  (string-match h host)))))))
      (parameterize ((current-urlencoded-arguments args)
		     (current-hostname host)
		     (current-path path))
	(if handler
	    (handler (lambda () (handle-entry (current-workdir) (normalize-path path))))
	    (begin
	      (log-error "No match for vhost ~A" host)
	      (old (current-request))))))))

(define (split-dir str)
  (let ((idx (string-index str #\/)))
    (if idx
	(values (string-take str idx) (string-drop str (add1 idx)))
	(values str #f))))

(define (apply-access-file path remaining-path continue)
  (let ((access-file (make-pathname path (spiffy-access-file))))
    (if (and (spiffy-access-file) (file-exists? access-file))
	((eval (call-with-input-file access-file read)) continue)
	(continue))))

(define (handle-entry incomplete-path remaining-path)
  (receive (file rest) (split-dir remaining-path)
    (let ((path (make-pathname (list (spiffy-root-path) incomplete-path) file)))
      (cond
       (((spiffy-access-denied?) incomplete-path file) (http:write-error-response 403 "Forbidden"))
       ((directory? path)
	(apply-access-file path rest
			   (lambda ()
			     (cond
			       ((not rest)
				(if (and (not (string-null? (current-path))) (string=? (string-take-right (current-path) 1) "/"))
				    (let ((index-page (find (lambda (ip)
							      (file-exists? (make-pathname path ip)))
							    (spiffy-index-pages))))
				      (if index-page
					  (handle-entry (make-pathname incomplete-path file) index-page)
					  ((spiffy-handle-directory) path)))
				    (let* ((path (string-append (make-pathname incomplete-path file) "/"))
					   (url (http:request-url (current-request)))
					   (get-args (and-let* ((idx (string-index url #\?)))
							       (string-drop url (add1 idx)))))
				     (if get-args
					 (moved-permanently (string-append path "?" get-args))
					 (moved-permanently path)))))
			       (else (handle-entry (make-pathname incomplete-path file) rest))))))
       ((file-exists? path)
	(let* ((ext (or (pathname-extension path) ""))
	       (unknown-handler (lambda _
				  (let ((ct (file-ext->mime-type ext)))
				    (set-header! (sprintf "Content-Type: ~a" ct))
				    ((content-transporter ct) ct path)))))
	  (parameterize ((current-pathinfo rest)
			 (current-workdir incomplete-path))
	   ;; XXX This interface is not too great.  If we want the full path we can
	   ;; always stick (current-workdir) in front of it.  Maybe just pass the file
	   ;; name?  Right now there's no easy way to find out the filename under web-root
	   ((alist-ref ext (spiffy-file-ext-handlers) string=? unknown-handler) path))))
       (else ((spiffy-not-found-handler)))))))


(define build-error-message
  (let ()
    (define exn-message (condition-property-accessor 'exn 'message "(no message)"))
    (define exn-location (condition-property-accessor 'exn 'location "(unknown location)"))
    (define exn-arguments (condition-property-accessor 'exn 'arguments '()))
    (define exn? (condition-predicate 'exn))
    (lambda (exn)
      (let ((chain (with-output-to-string print-call-chain)))
	(with-output-to-string
	  (lambda ()
	    (if (exn? exn)
		(begin
		  (display "<hr /><h2>Error:")
		  (and-let* ([loc (exn-location exn)])
			    (printf " (<em>~A</em>)" (htmlize (->string loc))))
		  (printf "</h2><h3>~A</h3>" (htmlize (exn-message exn)))
		  (unless (null? (exn-arguments exn))
			  (printf "<ul>")
			  (for-each
			   (lambda (a)
			     (##sys#with-print-length-limit 120 (lambda () (printf "<li>~S</li>" (htmlize (->string a))))))
			   (exn-arguments exn))
			  (printf "</ul>"))
		  (printf "<pre>~a</pre>" (htmlize chain)))
		(begin
		  (##sys#with-print-length-limit
		   120
		   (lambda ()
		     (printf "<hr /><strong>Uncaught exception:</strong> ~S" exn)))))))))))

(define (default-exn-handler exn)
  (write-fragment-response
   (display-error (car (current-response-code))
		  (cdr (current-response-code))
		  (build-error-message exn))))

;;; Cache:

(define spiffy-cache-limit 0)

(define cache (make-hash-table string=?))
(define cache-size 0)
(define lru-list '())

(define (cache-write path data)
  (let ([size (##sys#size data)])
    (cond [(zero? spiffy-cache-limit)]
	  [(> (+ cache-size size) spiffy-cache-limit)
	   (spiffy-debug "(cache) too large: ~S (size: ~A, current: ~A)" path size cache-size) ]
	  [(memq path lru-list) (spiffy-debug "(cache) already stored: ~S" path)]
	  [else
	   (spiffy-debug "(cache) stored: ~S (size: ~A, current: ~A)" path size cache-size)
	   (hash-table-set! cache path data)
	   (set! lru-list (cons path lru-list))
	   (set! cache-size (+ cache-size size))])))

(define (cache-read path)
  (and-let* ([data (hash-table-ref/default cache path #f)])
    (spiffy-debug "(cache) retrieved: ~S (size: ~A, current: ~A)" path (##sys#size data) cache-size)
    (when (or (null? lru-list) (not (string=? path (car lru-list))))
      (set! lru-list (cons path (delete path lru-list string=?))))
    data) )

(define (invalidate-cache . path)
  (if (pair? path)
      (begin
	(set! lru-list (delete path lru-list string=?))
	(hash-table-delete! cache (car path)))
      (set! cache (make-hash-table string=?))))


;;; Custom output procedure:
;
; - Kindly suggested by Kirill Lisovsky

(define (write-fragment-response . fragments)
  (let ([dest (make-string 256)]
	[dsize 256])
    (define (grow len n)
      (when (>= (+ n len) dsize)
	(let* ([dsize2 (+ dsize (max dsize n))]
	       [dest2 (make-string dsize2)] )
	  (##sys#copy-bytes dest dest2 0 0 len)
	  (set! dest dest2)
	  (set! dsize dsize2))))
    (let loop ([fragments fragments] [pos 0])
      (if (null? fragments) 
	  (begin
	    (set-header! (sprintf "Content-Length: ~A" pos))
	    (write-response-header)
	    (write-string dest pos) )
	  (let ([rest (cdr fragments)])
	    (match (car fragments)
	      [(or #f '()) (loop rest pos)]
	      [(x . y) (loop (cons x (cons y rest)) pos)]
	      [(? char? c)
	       (grow pos 1)
	       (string-set! dest pos c) 
	       (loop rest (add1 pos)) ]
	      [x (let* ([s (->string x)]
			[slen (string-length s)] )
		   (grow pos slen)
		   (##sys#copy-bytes s dest 0 pos slen) 
		   (loop rest (+ pos slen)))]))))))


;;; Other handy things:

(define (redirect-page loc)
  (current-response-code '(303 . "See Other"))
  (set-header! (sprintf "Location: ~A" loc))
  (write-fragment-response (sprintf "<html><head><title>Moved</title><body><h1>Moved</h1>Please see <a href=\"~A\">~A</a>.</body></html>" loc loc)))

(define (moved-permanently loc)
  (current-response-code '(301 . "Moved Permanently"))
  (set-header! (sprintf "Location: ~A" loc))
  (write-fragment-response (sprintf "<html><head><title>Moved permanently</title><body><h1>Moved permanently</h1>Please see <a href=\"~A\">~A</a>.</body></html>" loc loc)))

(define (redirect loc)
  (current-response-code '(302 . "Found"))
  (set-header! (sprintf "Location: ~A" loc)))

; This introduces spaces, which would otherwise be translated by
; htmlize-with-spaces into <br&nbsp;/>\n  etc
(define (htmlize-after str)
  (string-translate* str
		     '(("\n" . "<br />\n"))))

(define (htmlize-before str)
  (string-translate* str
		     '(("<" . "&lt;")
		       (">" . "&gt;")
		       ("\"" . "&quot;")
		       ("&" . "&amp;"))))

(define (htmlize-with-spaces str . args)
  (let-optionals args ([tabwidth 8])
    (let* ([space "&nbsp;"]
	   [tab (xsubstring space 0 (* (string-length space) tabwidth))])
      (htmlize-after
	(string-translate* (htmlize-before str)
			 `((" " . ,space)
			  ("\t" . ,tab)))))))

(define htmlize htmlize-before)

(define (strip-tags str)
  (string-substitute* str '(("<[A-Za-z0-9/:_]+>" . ""))) )

(define (escape-chars str)
  (with-output-to-string
    (lambda ()
      (let ([len (string-length str)])
	(let loop ([i 0] [start 0])
	  (if (>= i len)
	      (when (> i start)
		(display (substring str start len)) )
	      (let ([c (string-ref str i)])
		(if (or (char-alphabetic? c) (char-numeric? c))
		    (loop (add1 i) start)
		    (let ([n (char->integer c)])
		      (when (> i start)
			(display (substring str start i)) )
		      (write-char #\%)
		      (when (< n 16) (write-char #\0))
		      (display (number->string n 16)) 
		      (loop (add1 i) (add1 i)))))))))))

(define *header-regexp*
  (regexp
   (if (feature? 'pregexp)
       "([_A-Z0-9a-z---]+)\\:[ \t]*([^ \t].*)"
       "([-_A-Z0-9a-z]+)\\:[ \t]*([^ \t].*)") ) )

(define (single-header? name)
  (not (member name '("Set-Cookie") string-ci=?)))

(define (set-header! header)
  (match (string-match *header-regexp* header)
    [(_ name val)
     (spiffy-debug "set-header: ~S -> ~S" name val)
     (if (single-header? name)
	 (let* ([rh (current-response-headers)]
		[a (assoc name rh string-ci=?)] )
	   (if a
	       (set-cdr! a val)
	       (current-response-headers (alist-cons name val rh)) ) )
	 (current-response-headers (alist-cons name val (current-response-headers))))]
    [_ (error "invalid header syntax" header)]))

(define loaded-files '())

(define (load-once filename)
  (let* ([fn (make-pathname (list (spiffy-root-path) (current-workdir)) filename)]
	 [a (assoc fn loaded-files)]
	 [tm (file-modification-time fn)])
    (if a
	(and (> tm (cdr a))
 	       (begin
		 (load fn)
	         (set-cdr! a tm)
		 #t) )
  	(begin
	  (load fn)
	  (set! loaded-files (alist-cons fn tm loaded-files))
	  #t))))

;;; Custom error handler:

(define (display-error code msg . rest)
  (emit-error-message (sprintf "~a ~a" code msg) #f)
  (sprintf
   "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\"><html lang=\"en\"><head><title>Error!</title><style type=\"text/css\">~A</style></head><body><h1>~A ~A</h1><h3>Sorry, your request could not be handled.</h3>~A<hr /><p id=\"server-name\">~A at ~A, port ~A</p></body></html>"
   (spiffy-error-css) code msg (:optional rest "") (spiffy-server-name) (get-host-name) (spiffy-tcp-port)))

(http:error-response-handler
 display-error)

(define (stop-server) (exit 1))

;;; Main procedure:
(define (start-server #!key
		      (location (get-host-name))
		      (port (spiffy-tcp-port))
		      (root (spiffy-root-path))
		      (debug (spiffy-debug-mode))
		      (init noop) )
  ;; Now load config file, which may override all these things yet
  (parameterize ((load-verbose #f)  ;; We don't want "loading foo.so" ending up in output.
		 (spiffy-root-path root)
		 (spiffy-debug-mode debug)
		 (spiffy-tcp-port port))
    (when (spiffy-config-file)
          (load (spiffy-config-file)))
    ((http:make-server
      (spiffy-tcp-port)
      name: (spiffy-server-name)
      accept: (spiffy-accept-incoming-host)
      location: location
      init:
      (lambda ()
	(when (spiffy-user-name)
	      (let ([info (user-information (spiffy-user-name))])
		(unless info
			(error "user does not exist" (spiffy-user-name)) )
		(set-group-id! (list-ref info 3))
	       (initialize-groups (spiffy-user-name) (list-ref info 3))
	       (setenv "HOME" (list-ref info 5))
	       (set-user-id! (list-ref info 2)) ) )
	(when (eq? 'msvc (build-platform)) (http:force-close #t))
	(init)))
     (spiffy-debug-mode))))
